package fr.bhprod.sga.terrain.layer;

import java.util.ArrayList;

import org.apache.log4j.Logger;

import com.jme.bounding.BoundingBox;
import com.jme.image.Texture;
import com.jme.math.FastMath;
import com.jme.math.Matrix4f;
import com.jme.math.Ray;
import com.jme.math.Vector2f;
import com.jme.math.Vector3f;
import com.jme.renderer.ColorRGBA;
import com.jme.renderer.Renderer;
import com.jme.scene.TriMesh;
import com.jme.scene.state.AlphaState;
import com.jme.scene.state.TextureState;
import com.jme.util.TextureManager;
import com.jme.util.geom.BufferUtils;
import com.jmex.effects.ProjectedTextureUtil;

import fr.bhprod.sga.terrain.Terrain;
import fr.bhprod.sga.terrain.object.Decal;

/**
 * 
 * @author jbeaufils
 * @version 1.0
 */

public final class DecalUtil {

	public static final Logger LOGGER = Logger.getLogger(DecalUtil.class);

	/**
	 * Create a decal.<BR>
	 * The rotation of the decal is changed with the up vector
	 * and his size is changed with the distance of the projector
	 * @param display The current display
	 */
	public static final Decal createNewDecal(Vector3f position, String decalName, String decalFile, Terrain terrain, Renderer renderer) {
		Vector3f projectorPos = new Vector3f(position.x, 100f, position.z);
		Vector3f up = Vector3f.UNIT_X;
		float near = 1.0f;
		float far = 100.0f;
		float fov = 20f;
		float aspect = 1.0f;
		
		Texture texture = TextureManager.loadTexture(decalFile,
				Texture.MM_LINEAR_LINEAR, Texture.FM_LINEAR);
		
		Decal decal = new Decal();
		decal.setName(decalName);
		decal.setUp(up);
		decal.setProjectorPos(projectorPos);
		decal.setTexture(texture);
		decal.setNear(near);
		decal.setFar(far);
		decal.setFov(fov);
		decal.setAspect(aspect);
		
		putDecal(decal, terrain, renderer);
		
		return decal;
	}

	/**
	 * Draw a decal.
	 * @param display The current display
	 */
	public static final void putDecal(Decal decal, Terrain terrain, Renderer renderer) {
		float h = FastMath.tan( decal.getFov() * FastMath.DEG_TO_RAD ) * decal.getFar() * .5f;
		float w = h * decal.getAspect();

		Matrix4f matRot = new Matrix4f();
		ProjectedTextureUtil.matrixLookAt(decal.getProjectorPos(), decal.getLookAt(), decal.getUp(), matRot);

		// Calculate the rays of the borders of the frustum
		Vector3f topLeft = new Vector3f(-w,h,-decal.getFar());
		topLeft = matRot.mult(topLeft).add(decal.getProjectorPos());
		Ray rTopLeft = new Ray(decal.getProjectorPos(), topLeft.subtract(decal.getProjectorPos()).normalize());

		Vector3f topRight = new Vector3f(w,h,-decal.getFar());
		topRight = matRot.mult(topRight).add(decal.getProjectorPos());
		Ray rTopRight = new Ray(decal.getProjectorPos(), topRight.subtract(decal.getProjectorPos()).normalize());

		Vector3f bottomLeft = new Vector3f(-w,-h,-decal.getFar());
		bottomLeft = matRot.mult(bottomLeft).add(decal.getProjectorPos());
		Ray rBottomLeft = new Ray(decal.getProjectorPos(), bottomLeft.subtract(decal.getProjectorPos()).normalize());

		Vector3f bottomRight = new Vector3f(w,-h,-decal.getFar());
		bottomRight = matRot.mult(bottomRight).add(decal.getProjectorPos());
		Ray rBottomRight = new Ray(decal.getProjectorPos(), bottomRight.subtract(decal.getProjectorPos()).normalize());

		
		// Search the intersection between the four rays and the terrain
		float xMin = Float.MAX_VALUE;
		float xMax = -Float.MAX_VALUE;
		float zMin = Float.MAX_VALUE;
		float zMax = -Float.MAX_VALUE;

		for (int tz = 0; tz < terrain.getTilesZ(); tz += 16) {
			for (int tx = 0; tx < terrain.getTilesX(); tx += 16) {
				TriMesh m = (TriMesh)terrain.getLayerTextureDetailNode().getChild(
						"TextureDetailLayer[" + (tx / 16) + "][" + (tz / 16) + "]");

				// Test topLeft
				ArrayList<Integer> pickList = new ArrayList<Integer>();
				m.findTrianglePick(rTopLeft, pickList, 0);
				if (!pickList.isEmpty()) {
					Vector3f[] tri = new Vector3f[3];
					m.getTriangle(pickList.get(0).intValue(), tri);
					for (int i = 0; i < 3; i++) {
						if (tri[i].x < xMin) {
							xMin = tri[i].x;
						}
						if (tri[i].x > xMax) {
							xMax = tri[i].x;
						}
						if (tri[i].z < zMin) {
							zMin = tri[i].z;
						}
						if (tri[i].z > zMax) {
							zMax = tri[i].z;
						}
					}
				}

				// Test topRight
				pickList = new ArrayList<Integer>();
				m.findTrianglePick(rTopRight, pickList, 0);
				if (!pickList.isEmpty()) {
					Vector3f[] tri = new Vector3f[3];
					m.getTriangle(pickList.get(0).intValue(), tri);
					for (int i = 0; i < 3; i++) {
						if (tri[i].x < xMin) {
							xMin = tri[i].x;
						}
						if (tri[i].x > xMax) {
							xMax = tri[i].x;
						}
						if (tri[i].z < zMin) {
							zMin = tri[i].z;
						}
						if (tri[i].z > zMax) {
							zMax = tri[i].z;
						}
					}
				}

				// Test bottomLeft
				pickList = new ArrayList<Integer>();
				m.findTrianglePick(rBottomLeft, pickList, 0);
				if (!pickList.isEmpty()) {
					Vector3f[] tri = new Vector3f[3];
					m.getTriangle(pickList.get(0).intValue(), tri);
					for (int i = 0; i < 3; i++) {
						if (tri[i].x < xMin) {
							xMin = tri[i].x;
						}
						if (tri[i].x > xMax) {
							xMax = tri[i].x;
						}
						if (tri[i].z < zMin) {
							zMin = tri[i].z;
						}
						if (tri[i].z > zMax) {
							zMax = tri[i].z;
						}
					}
				}

				// Test bottomRight
				pickList = new ArrayList<Integer>();
				m.findTrianglePick(rBottomRight, pickList, 0);
				if (!pickList.isEmpty()) {
					Vector3f[] tri = new Vector3f[3];
					m.getTriangle(pickList.get(0).intValue(), tri);
					for (int i = 0; i < 3; i++) {
						if (tri[i].x < xMin) {
							xMin = tri[i].x;
						}
						if (tri[i].x > xMax) {
							xMax = tri[i].x;
						}
						if (tri[i].z < zMin) {
							zMin = tri[i].z;
						}
						if (tri[i].z > zMax) {
							zMax = tri[i].z;
						}
					}
				}
			}
		}

		
		// Create the mesh for the projection
		int nbTilesX = (int)((xMax - xMin) / terrain.getTileRealSize() + 1);
		int xTileMin = (int)(xMin / terrain.getTileRealSize())+1;
		int nbTilesZ = (int)((zMax - zMin) / terrain.getTileRealSize() + 1);
		int zTileMin = (int)(-zMax / terrain.getTileRealSize())+1;

		int nbSquare = nbTilesX * nbTilesZ;
		
		Vector3f[] vertexes = new Vector3f[nbSquare];
		Vector3f[] normals = new Vector3f[nbSquare];
		Vector2f[] texCoords = new Vector2f[nbSquare];
		ColorRGBA[] colors = new ColorRGBA[nbSquare];
		
		// For each vertex
		for (int z = 0; z < nbTilesZ; z++) {
			for (int x = 0; x < nbTilesX; x++) {
				vertexes[z * nbTilesX + x] = new Vector3f((xTileMin + x) * terrain.getTileRealSize(),
						terrain.getHeights()[xTileMin + x - 1][zTileMin + z - 1] + 1f, (-zTileMin - z) * terrain.getTileRealSize());
				normals[z * nbTilesX + x] = new Vector3f(0, 0, 1);
				texCoords[z * nbTilesX + x] = new Vector2f(
						(float) x/2, (float) z/2);
				colors[z * nbTilesX + x] = new ColorRGBA(0,0,0,0);
			}
		}

		// 2 tris for each tile * 3 vertices
		int[] indexes = new int[(nbTilesX-1) * (nbTilesZ-1) * 2 * 3];

		// For each tile
		for (int z = 0; z < (nbTilesZ-1); z++) {
			for (int x = 0; x < (nbTilesX-1); x++) {
				int i = z * ((nbTilesX-1) * 6) + (x * 6);
				indexes[i] = z * nbTilesX + x;
				indexes[i + 1] = z * nbTilesX + x + 1;
				indexes[i + 2] = (z + 1) * nbTilesX + x;
				indexes[i + 3] = z * nbTilesX + x + 1;
				indexes[i + 4] = (z + 1) * nbTilesX + x + 1;
				indexes[i + 5] = (z + 1) * nbTilesX + x;
			}
		}

		TriMesh mesh = new TriMesh(decal.getName());

		TextureState ts = renderer.createTextureState();
		ts.setTexture(decal.getTexture());

		// Set the texture up for projection
		ProjectedTextureUtil.setupProjectedTexture(decal.getTexture(),
				Texture.WM_BCLAMP_S_BCLAMP_T, Texture.ACF_ADD);

		mesh.setRenderState(ts);

		AlphaState as = renderer.createAlphaState();
		as.setBlendEnabled(true);
		as.setDstFunction(AlphaState.DB_ONE_MINUS_SRC_ALPHA);
		as.setSrcFunction(AlphaState.SB_SRC_ALPHA);
		mesh.setRenderState(as);

		mesh.reconstruct(BufferUtils.createFloatBuffer(vertexes), BufferUtils
				.createFloatBuffer(normals), BufferUtils
				.createFloatBuffer(colors), BufferUtils
				.createFloatBuffer(texCoords), BufferUtils
				.createIntBuffer(indexes));

		mesh.setModelBound(new BoundingBox());
		mesh.updateModelBound();

		// Update decal
		ProjectedTextureUtil.updateProjectedTexture(decal.getTexture(), decal.getFov(), decal.getAspect(), decal.getNear(), decal.getFar(), decal.getProjectorPos(), decal.getLookAt(), decal.getUp());
		
		terrain.getLayerDecalsNode().attachChild(mesh);

		terrain.getLayerDecalsNode().updateRenderState();
		terrain.getLayerDecalsNode().updateGeometricState(0, true);
		
		decal.setMesh(mesh);
	}

}

